#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "dtypes.h"
#include "opcodes.h"

#define MAXL 100000

struct {
   uint16 dir;
   char name[64];
} labtab[MAXL];
int nlabel=0;
FILE *fpo,*fplst;

int checknum(char *p)
{
    int i;
    // Primer caracter debe ser [0-9]
    if (*p<'0' || *p>'9') return -1;
    // Si "0X.." entonces es hexadecimal
    if (p[0]=='0' && p[1]=='X') {
    	p=&p[2]; i=0;
	while ((*p>='0' && *p<='9') || (*p>='A' && *p<='F')) {
	    i<<=4;
	    if (*p>='0' && *p<='9') i+=*p-'0';
	    else i+=*p-'A'+10;
	    p++;
	}
	return i;
    } else return atoi(p);
}

int checknemo(char *p)
{
    int i;
    for (i=0;i<16;i++) {
       if (strcmp(instr[i],p)==0) return i;
    }
    return 0xff;
}

int checkcc(char *p)
{
    int i;
    for (i=0;i<8;i++) {
       if (strcmp(ccond[i],p)==0) return i;
    }
    return 0xff;
}

int checkreg(char *p,int pointer)
{
    int i;
    char buf[8];
    const char *reg[]={ "R0","R1","R2","R3","R4","R5","R6","R7",    			
    			"PC","SP","LR"}; // Registros alias
    for (i=0;i<11;i++) {
       sprintf(buf,"(%s)",reg[i]);
       if (pointer==0 && strcmp(reg[i],p)==0) return i&7;
       if (pointer==1 && strcmp(buf,p)==0) return i&7;
    }
    return 0xff;
}

char delim,tok[256];
char *ptok;

int scantok()
{
    const char *delim1=" \t";
    const char *delim2="\n; \t:.,";
    char *pp;
    
    while (strchr(delim1,*ptok) && *ptok) ptok++;
    pp=ptok;
    while (!strchr(delim2,*ptok) && *ptok) ptok++;
    delim=*ptok;
    *ptok++=0;
    strcpy(tok,pp);
    //printf(">%s< delim:>%c<\n",tok,delim);
}

// Buscamos los valores de las etiquetas
void pass1(char *fname)
{
    FILE *fp;
    int cdir=0,i,op;
    char *p,*pp,buf[256];
    
    if ((fp=fopen(fname,"r"))==NULL) exit(0);
    
    for (nlabel=0;;) {
        fgets(buf,255,fp);
	if (feof(fp)) break;
	p=buf;
	while (*p) {*p=toupper(*p); p++;}	
	//printf("%s",buf);
	ptok=buf;
uno:
	scantok();	
	if (strlen(tok)==0) continue;
	if (delim==':') {
	    // etiqueta
	    strcpy(labtab[nlabel].name,tok);
	    labtab[nlabel++].dir=cdir;
	    goto uno; // seguimos buscando mnemonico
	}
	 
	if(strcmp(tok,"ORG")==0) {
	    scantok();
	    i=checknum(tok);
	    if (i==-1) {
		printf("Error: valor incorrecto: %s\n",tok);
		exit(1);
	    }
	    cdir=i;
	    continue;
	}
	if(strcmp(tok,"WORD")==0){
	    cdir++;
	    continue;
	} 
	if(checknemo(tok)<16 || strcmp(tok,"NOP")==0) {
	    cdir++;
	    continue;
	}
		
	printf("Error: >%s<\n",tok);
	exit(1);

    }
    fclose(fp);
    fprintf(fplst,"Lista de etiquetas\n");
    fprintf(fplst,"------------------------------------------------------\n");
    for (i=0;i<nlabel;i++) {
         fprintf(fplst,"%04X  %s\n",labtab[i].dir,labtab[i].name);
    }
    fprintf(fplst,"------------------------------------------------------\n");

}

// Aqui se ensambla

int etival(char *peti)
{
   int i;
   for (i=0;i<nlabel;i++) {
       if (strcmp(labtab[i].name,peti)==0) return labtab[i].dir;
   }
   return -1;
}

void emit(int dir, int dato, char *orig)
{
    static int bdir=0;
    int i,j;

    if (dir>=0) fprintf(fplst,"%04X  -  %04X   %s",dir,dato,orig);

    if (dir==bdir+1) {
        bdir=dir;
	fprintf(fpo,"%04X\n",dato);
    }else {
        bdir=dir;
        fprintf(fpo,"@%04X\n",bdir);
        fprintf(fpo,"%04X\n",dato);
    }
}

void pass2(char *fname)
{
    FILE *fp,*fpo;
    int cdir=0,i,op,cc,ra,rb,rd,n,disp,nl,dumpfl;
    char *p,*pp,buf[256],bf[256];
    
    if ((fp=fopen(fname,"r"))==NULL) exit(0);
  
    for (nl=0;;) {
        fgets(buf,255,fp);
	nl++;
	//printf(">%s<\n",buf);
	if (feof(fp)) break;
	dumpfl=0;
	p=buf;
	while (*p) {*p=toupper(*p); p++;}
	strcpy(bf,buf);
	ptok=bf;
dos:
	scantok();
	cc=0;
	switch(delim){
	  case ':':	// etiqueta: ignorar en esta fase
	  	dumpfl=1;
	  	goto dos;
	  case '\n':    
	  case ';':
	  case 0:
		if (strcmp(tok,"NOP")==0) {
		    emit(cdir,0,buf);
		    cdir++;
		    break;	    
		}
		if (strcmp(tok,"RETI")==0) {
		    emit(cdir,(RETI<<12)|(1<<9),buf);
		    cdir++;
		    break;	    
		}
		if ((op=checknemo(tok))<15) {
		    printf("(%d) Error, Pocos operandos: %s\n",nl,tok);
		    exit(1);
		}
		if (dumpfl) fprintf(fplst,"%04X\t\t%s",cdir,buf);
		continue;	  
	  case '.':  cc=1;  // instr. condicional
	  case ' ':
	  case '\t':	// instruccion/directiva
	        if (strcmp(tok,"ORG")==0) {
		    scantok();
		    i=checknum(tok);
		    if (i==-1) {
			printf("(%d) Error: valor invalido: %s\n",nl,tok);
			exit(1);
		    }
		    cdir=i;
		    break;
		}
	  	if (strcmp(tok,"WORD")==0) {
		    scantok();
		    if ((i=etival(tok))>=0) {
		        emit(cdir,i,buf);
		    } else {
		        i=checknum(tok);
			if (i==-1) {
			    printf("(%d) Error: Etiqueta desconocida: %s\n",nl,tok);
			    exit(1);
			}
		        emit(cdir,i,buf);
		    }
		    //printf("\n");
		    cdir++;
		    break;	    
		}
		if (strcmp(tok,"NOP")==0) {
		    emit(cdir,0,buf);
		    //printf("\n");
		    cdir++;
		    break;	    
		}
		
		op=checknemo(tok);
		//printf("op=%d\n",op);
		if (op>15) {
		    printf("(%d) Error, Mnemonico desconocido: %s\n",nl,tok);
		    exit(1);
		} else {
		    if (cc) {
		        scantok();
			if ((cc=checkcc(tok))>=8) {
			   printf("(%d) Error, mal codigo de condicion: %s\n",nl,tok);
			   exit(1);
			}
		    } else cc=1;
		    switch(op) {
		    // 3 operandos en registros
		    case ADD:
		    case SUB:
		    case ADC:
		    case SBB:
		    case AND:
		    case OR:
		    case XOR:
			scantok();
			if ((ra=checkreg(tok,0))>7) {
			   printf("(%d) Error, mal registro: %s\n",nl,tok);
			   exit(1); 
			}
			scantok();
			if ((rb=checkreg(tok,0))>7) {
			   printf("(%d) Error, mal registro: %s\n",nl,tok); 
			   exit(1);
			}
			scantok();
			if ((rd=checkreg(tok,0))>7) {
			   printf("(%d) Error, mal registro: %s\n",nl,tok);
			   exit(1); 
			}
			emit(cdir,(op<<12)|(cc<<9)|(rd<<6)|(ra<<3)|rb,buf);
			break;
		    case NEG:
		    case NOT:
		    case ROR:
		        ra=0;
		        scantok();
			if ((rb=checkreg(tok,0))>7) {
			   printf("(%d) Error, mal registro: %s\n",nl,tok); 
			   exit(1);
			}
			scantok();
			if ((rd=checkreg(tok,0))>7) {
			   printf("(%d) Error, mal registro: %s\n",nl,tok);
			   exit(1); 
			}
			emit(cdir,(op<<12)|(cc<<9)|(rd<<6)|(ra<<3)|rb,buf);
			break;
		    case ADDQ:
		    case SUBQ:
		        scantok();
			if ((ra=checkreg(tok,0))>7) {
			   printf("(%d) Error, mal registro: %s\n",nl,tok);
			   exit(1); 
			}
			scantok();
			n=atoi(tok); n&=7;
			scantok();
			if ((rd=checkreg(tok,0))>7) {
			   printf("(%d) Error, mal registro: %s\n",nl,tok);
			   exit(1); 
			}
			emit(cdir,(op<<12)|(cc<<9)|(rd<<6)|(ra<<3)|n,buf);
			break;
		    case LD:
		        scantok();
			if ((rb=checkreg(tok,1))>7) {
			   printf("(%d) Error, mal registro: %s\n",nl,tok);
			   exit(1);
			}
			scantok();
			if ((rd=checkreg(tok,0))>7) {
			   printf("(%d) Error, mal registro: %s\n",nl,tok);
			   exit(1); 
			}
			emit(cdir,(op<<12)|(cc<<9)|(rd<<6)|rb,buf);
			break;
		    case ST:
		    	scantok();
			if ((ra=checkreg(tok,0))>7) {
			   printf("(%d) Error, mal registro: %s\n",nl,tok);
			   exit(1);
			}
			scantok();
			if ((rb=checkreg(tok,1))>7) {
			   printf("Error, mal registro: %s\n",tok);
			   exit(1); 
			}
			emit(cdir,(op<<12)|(cc<<9)|(ra<<3)|rb,buf);
		        break;
		    case JR:
		        scantok();
			if ((i=etival(tok))>=0) {
			   disp=i-cdir-1;
		        } else {
			    i=checknum(tok);
			    if (disp==-1) {
			        printf("(%d) Error: Etiqueta desconocida: %s\n",nl,tok);
			        exit(1);
			    }
			    disp=i-cdir-1;
		        }
			if (disp>255 || disp<-256){
			    printf("(%d) Error: Desplaz. fuera de rango\n",nl,tok);
			    exit(1);
			}
			disp&=511;
			emit(cdir,(op<<12)|(cc<<9)|disp,buf);
			break;
		    case RETI:
		        emit(cdir,(op<<12)|(cc<<9),buf);
			break;
		    }
		    cdir++;
		}
	}
    }
    fclose (fp);
    
    //emit(-1,0,"\n"); // Vaciamos el buffer
}
 
main(int argc, char **argv)
{
     int i,w,n;
     uint16 dir,op;
     FILE *fp;
 
     if ((fpo=fopen("out.hex","w"))==NULL) exit(0);
     if ((fplst=fopen("out.lst","w"))==NULL) exit(0);
     pass1(argv[1]);
     pass2(argv[1]);
     fclose(fpo);
     fclose(fplst);
     printf("OK     object: out.hex   Listing: out.lst\n");
}
