#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>

#include "dtypes.h"
#include "opcodes.h"

const char *rgn0[8]={"PC","R1","R2","R3","R4","R5","R6","R7"};
const char *rgn1[8]={"PC","SP","LR","R3","R4","R5","R6","R7"};
char **rgn=rgn1;
int Trace;
int Trap=-1;
int spreg=1;

struct {
    /* Registros */
    int32 r[8];	/* Banco de registros PC=r[0] */
    int32 r0i;	/* PC durante interrupciones */
    uint16 ir;		/* Registro de instruccin */

    /* FLAGS: Modo normal y modo interrupcion*/
    uint8 flags,flagsi;
    
    uint8 intff;	/* Flag de deteccion de interrupcion */
    uint8 modo;	/* modo de funcionamiento 0: normal 1:Int */
    /* Otras variables */
    uint32 nciclos;	/* Contador de ciclos de reloj */
    uint32 ninstrs; 	/* Contador de instrucciones */
    uint32 nnops;	/* Contador de NOPs de programa */
    
} cpu;

uint16 MEM[0x10000];
int dirtyvideo=0;

/*--------------------------------------------------------------------*/

void reset_cpu();

/*--------------------------------------------------------------------*/
videorefresh()
{
    int i,j,dir;
    unsigned char a;

    puts("\033[?25l");
    printf("\033[2J\033[H");
    for (i=0,dir=0x1000;i<24;i++) {
	for (j=0;j<80;j++) {
	    a=MEM[dir++];
	    putchar(isprint(a)?a:'.');
	}
	if (i<23)putchar('\n');
    }
    fflush(stdout);
}

int muestra_instr(uint16 cop, uint16 pc)
{
	uint8 ra,rb,rd,cc,op,n;
	int16 desp;
	int i;

	n=rb=cop & 7;  
	desp=cop & 0x1ff; if (desp & 0x100) desp |=0xfe00;
	ra=(cop>>3) & 7;
	rd=(cop>>6) & 7;
	cc=(cop>>9) & 7;
	op=(cop>>12) & 15;
	
	if (!cc) { printf("NOP"); return 0; }
	printf("%s",instr[op]);
	i=strlen(instr[op]);
	if (cc!=1) {printf(".%s",ccond[cc]); i+=strlen(ccond[cc])+1;}
	for (;i<8;i++) putchar(' ');
	switch(op) {
	case ADD:
	case SUB:
	case ADC:
	case SBB:
	case AND:
	case OR:
	case XOR:
		printf("%s,%s,%s",rgn[ra],rgn[rb],rgn[rd]);
		break;
	case ADDQ:
	case SUBQ:
		printf("%s,%d,%s",rgn[ra],rb,rgn[rd]);
		break;
	case NOT:
	case NEG:
	case ROR:
		printf("%s,%s",rgn[rb],rgn[rd]);
		break;
	case LD:
		printf("(%s),%s",rgn[rb],rgn[rd]);
		if (!rb) return 1;
		break;
	case ST:
		printf("%s,(%s)",rgn[ra],rgn[rb]);
		break;
	case JR:
		if (desp>=0) putchar('+');
		printf("%d",desp);
		i=0; n=abs(desp); while (n) {i++; n/=10;}
		for(;i<7;i++) putchar(' ');
		printf(";0x%04X",pc+desp);

		break;
	case RETI:
		break;
	}
	return 0;
}

/*--------------------------------------------------------------------*/
extern inline int debug(uint16 fetched_ir)
{
    int i,j,k,ldpc;
    uint16 dir;
    uint32 *ppc;
    const char *fln="SZC";
    char buf[64];
    
    ppc=(cpu.modo)?&cpu.r0i:&cpu.r[0];

dupdate:
    
    if (dirtyvideo) {
	videorefresh();
	printf("\033[24;1H\033[7m<break> press ENTER\033[0m");

	fflush(stdout);
        getchar();
	puts("\033[?25h");
	dirtyvideo=0;
    }
    // borrar pantalla
    printf("\033[2J\033[H");
    
    // mostrar codigo actual
    printf("     Exec:   "); ldpc=muestra_instr(cpu.ir,*ppc);
    printf("\033[1;42HFetch: "); muestra_instr(fetched_ir,*ppc+1);
    printf("\n\n");
    for (i=0,dir=*ppc;i<16;i++,dir++) {
        printf("%04X: %04X   ",dir,MEM[dir]);
	if (ldpc) {
	    printf("word %04X",MEM[dir]);
	    ldpc=0;
	}else ldpc=muestra_instr(MEM[dir],dir+1);
	putchar('\n');
    }
    
    // Mostrar estado de CPU (registros, modo)
    printf("\033[2;42HMode=%s",(cpu.modo)?"Interrupt":"Normal");
    for (i=0;i<8;i++) {
    	printf("\033[%d;42H%s=",i+4,rgn[i]);
	if (cpu.r[i]<0) printf("----");
	else printf("%04X",cpu.r[i]);
	if (!i) printf("    PCInt=%04X",cpu.r0i);
    }
    
    printf("\033[13;42HFlags=");
    j=cpu.flags;
    for (i=0;i<3;i++) {putchar((j&4)?fln[i]:'.'); j<<=1;}
    printf("   FlgInt=");
    j=cpu.flagsi;
    for (i=0;i<3;i++) {putchar((j&4)?fln[i]:'.'); j<<=1;}
    
    // Mostrar estadisticas, breakpoint
    printf("\033[15;42HCycles=%d",cpu.nciclos);
    printf("\033[16;42HIntr.s=%d",cpu.ninstrs);
    printf("\033[17;42HNOPs  =%d",cpu.nnops);
    if (Trap>=0) printf("\033[18;42HBreak =%04X",Trap);
    
    // Mostrar Pila
    printf("\033[3;66HStack");
    printf("\033[4;66H(R%d)",spreg);
    k=cpu.r[spreg]; if (k<=0) k=0x10000;
    j=k;
    k+=14; if (k>0xffff) k=0x10000;
    for (i=4+k-j;j<k;i--,j++) {
    	printf("\033[%d;66H%04X",i,MEM[j&0xffff]);
    }
    
    printf("\033[21;1H");
    
    // interprete de comandos
    while(1) {
    	printf("[Command,'?']-> ");
    	fflush(stdout);fflush(stdin);

    	fgets(buf,63,stdin);
    	for(j=0;buf[j]>=' ';j++) buf[j]=toupper(buf[j]);
	buf[j]='\0';
    	switch(buf[0]) {
	case '\0': return;			// paso simple
	case '?':  printf("\n?        This help\n");
		   printf("B <addr> Breakpoint at <addr>\n");
		   printf("= <addr> Exec until <addr>\n");
		   printf("N        Exec until Next instruction\n");
		   printf("J <addr> Exec from <addr> (Jump)\n");
		   printf("C        Continue execution without Breakpoint\n");
		   printf("E        Continue execution with Breakpoint\n");
		   printf("M [addr] Memory dump from [addr]\n");
		   printf("M        Memory dump from next address\n");
		   printf("D [addr] Disassembly from [addr]\n");
		   printf("D        Disassembly from next address\n");
		   printf("R        Reset state\n");
		   printf("Q        Quit to O.S.\n\n");
		   break;
	case 'B':  if(strlen(buf)>=2) {		// set BRK
      		     sscanf(buf+1,"%hX",&i);
		     Trap=i;
		   }
		   goto dupdate;
	case '=':  if(strlen(buf)>=2) {		// Ejecutar hasta BRK
      		     sscanf(buf+1,"%hX",&i);
		     Trap=i;
		     Trace=0;
		     videorefresh();
		     return 0;
		   }
                   break;
	case 'N':  Trap=(*ppc)+1;		// Ejecutar hasta PC+1
      		   Trace=0;
		   videorefresh();
		   return 0;

      	case 'J':  if(strlen(buf)>=2){		// Cambia PC y ejecuta
      		     sscanf(buf+1,"%hX",&i);
		     *ppc=i;
		     Trace=0;
		     videorefresh();
		     return 1;
		   }
                   break;
      	case 'C':  Trap=-1;			// Continua sin BRK
      	case 'E':  Trace=0;			// Continua con BRK
		   videorefresh();
		   return;
	case 'R':  reset_cpu();			// RESET
		   for (i=0;i<80*24;i++) MEM[i+0x1000]=' ';  // CLS
		   return 1;
      	case 'Q':  exit(0);			// Salida del emulador
	
	case 'M':  if(strlen(buf)>=2) {		// Listado de Memoria
		     sscanf(buf+1,"%hX",&dir);
		   }
		   for (i=0;i<16;i++) {
		     printf("%04X - ",dir);
		     for (j=0;j<8;j++) {
		       k=MEM[(dir+j)&0xffff];
		       printf("%04X ",k);
		     }
		     for (j=0;j<8;j++) {
		       k=MEM[(dir+j)&0xffff];
		       putchar((k>=32 && k<128)?k:'.');
		     }
		     putchar('\n');
		     dir+=j;
		   }
		   break;
	case 'D':  if(strlen(buf)>=2) {		// Desensamblado
		     sscanf(buf+1,"%hX",&dir);
		   }
		   for (i=0;i<16;i++,dir++) {
        	     printf("%04X: %04X   ",dir,MEM[dir]);
		     if (ldpc) {
			printf("word %04X",MEM[dir]);
			ldpc=0;
		     }else ldpc=muestra_instr(MEM[dir],dir);
		     putchar('\n');
		   }
		   break;

	default:   goto dupdate;
	}
    }
}

/**************----------- EMULADOR ----------*************/

uint16 rd_mem(uint16 dir)
{
	return MEM[dir];
}

void wr_mem(uint16 dir,uint16 dato)
{
    int x,y;
    unsigned char c;
	MEM[dir]=dato;
	// Emulamos framebuffer
	if (dir>=0x1000 && dir <(0x1000+2000)) {
	     y=(dir-0x1000)/80;
	     x=(dir-0x1000)%80;
	     printf("\033[%d;%dH%c",y+1,x+1,dato&0xff);
	     fflush(stdout);
	     dirtyvideo=1;
	}
}

/*--------------------------------------------------------------------*/

extern inline void un_ciclo(int interr)
{
    uint32 *ppc;
    uint8 *pfl;
    uint32 t;
    uint16 a,b,f,ir;
    uint8  ra,rb,rd,cc,op,n,exe,ls,c,z,s,intnop,nmodo,reti,wrpc,wr;
    uint16 desp,bus_a,bus_b,bus_d,bus_dir,bus_datos;
    
    /****** etapa DECODE/EXECUTE ******/
    
    n=rb=cpu.ir & 7;
    desp=cpu.ir & 0x1ff; if (desp & 0x100) desp |=0xfe00;
    ra=(cpu.ir>>3) & 7;
    rd=(cpu.ir>>6) & 7;
    cc=(cpu.ir>>9) & 7;
    op=(cpu.ir>>12) & 15;
    
    ls=0;	/* No Load-Store */
    wr=0;	/* No write (ST) */
    if (cpu.modo) {ppc=&cpu.r0i; pfl=&cpu.flagsi;}
    else {ppc=&cpu.r[0]; pfl=&cpu.flags;}
    c=(*pfl)&1;
    z=((*pfl)>>1)&1;
    s=((*pfl)>>2)&1;
    
    /****** Cdigos de condicin ******/
    
    switch(cc) {
	case 0:	exe=0; break;
	case 1: exe=1; break;
	case 2:	exe=c^1; break;
	case 3: exe=c; break; 
	case 4: exe=z^1; break;
	case 5: exe=z; break;
	case 6: exe=s^1; break;
	case 7: exe=s; break;
    }

    if (exe) {
    
	/****** buses ******/
	
	if (op==ADDQ || op==SUBQ) bus_b=n;
	else if (op==JR) bus_b=desp;
	else bus_b=(cpu.modo && rb==0)?cpu.r0i:cpu.r[rb];
	
	if (op==LD || op==ST) {
		bus_dir=bus_b;
		ls=1;
	}

	if (op==JR) bus_a=*ppc;
	else if (op==LD) bus_datos=bus_a=rd_mem(bus_dir);
	else bus_a=(cpu.modo && ra==0)?cpu.r0i:cpu.r[ra];
	if (op==ST) {
		bus_datos=bus_a;
		wr=1;
		//wr_mem(bus_dir,bus_datos);
	}
	
	/****** Operacin en la ALU ******/
	
	a=bus_a; b=bus_b;
	switch(op) {
	case ADD:
	case ADDQ:	t=a+b;		
			if(rd) c=t>>16;
			f=t&0xffff;
			break;
	case SUB:
	case SUBQ:	b=~b;		
			t=a+b+1;
			if(rd) c=(t>>16)^1;
			f=t&0xffff;
			break;
	case ADC:	t=a+b+(*pfl&1);	
			if(rd) c=t>>16;
			f=t&0xffff;
			break;
	case SBB:	b=~b;		
			t=a+b+((*pfl&1)^1);
			if(rd) c=(t>>16)^1;
			f=t&0xffff;
			break;
	case AND:	f=a&b; break;	
	case OR:	f=a|b; break;	
	case XOR:	f=a^b; break;	
	case NEG:	t=(~b)+1;	
			f=t&0xffff;
			break;
	case NOT: 	f=~b; break; 
	case ROR:	t=b; 
			if (*pfl&1) t|=0x10000;
			if(rd) c=t&1;
			f=t>>1;
			break;
	case LD:	f=a; break;
	case JR:	t=a+b;
			f=t&0xffff;
			break;	
	default:	f=0;
	}
	
	if (op<ST && rd>0) {
		if (f) z=0; else z=1;
		s=f>>15;
	}
	
	bus_d=f;
	
    }

    /****** variables utiles ******/
    wrpc=(exe && (op==JR || (op!=ST && op!=RETI && rd==0)));
    reti=(exe && op==RETI);
    	
    nmodo=(!reti &&(cpu.modo || (!wrpc && cpu.intff)));
    intnop=(!cpu.modo && !wrpc && cpu.intff);


    /****** Etapa de Fetch ******/
    t=0;
    if (ls || intnop) ir=bus_datos&0xf1ff;	/* insertamos NOP */
    else {
    	bus_dir=*ppc;
    	ir=bus_datos=rd_mem(bus_dir);
	t=1;
	if (bus_dir==Trap) Trace=1;	/* Breackpoint */
    }
    
    /* Este es el sitio para echar un vistazo a los buses y registros*/
    if (Trace) if (debug(ir)) return; // si se ha cambiado PC abortamos
    
    if (t) cpu.ninstrs++;
    if (t && !(ir&0x0e00)) cpu.nnops++;
    
    /****** y finalmente el flanco de reloj: actualizar registros ******/
    if (wr) wr_mem(bus_dir,bus_datos);		// memory write (ST)
    if (!ls && !intnop) (*ppc)++;
    else if (ls && exe && rb==0) (*ppc)++;

    wrpc=0;
    if (exe) {
    	if (op==JR) {*ppc=bus_d; wrpc=1;}
    	else if (op!=ST && op!=RETI) {
	    if (cpu.modo && rd==0) cpu.r0i=bus_d;
	    else cpu.r[rd]=bus_d;
	    if (rd==0) wrpc=1;
	}
	// Actualizamos flags
	*pfl=(s<<2)|(z<<1)|c;
    }
    
    /**** Interrupcion y modo ****/
    
    cpu.modo=nmodo;
    if (!cpu.modo) cpu.r0i=2;
    cpu.intff=(interr && !reti);
    
    cpu.ir=ir;	/* nueva instruccin al pipeline */
    
    /***** Y contamos el ciclo ******/
    cpu.nciclos++;
}

/*--------------------------------------------------------------------*/

void reset_cpu()
{
    int i;
    cpu.intff=0;	/* Sin interrupciones */
    cpu.modo=0;		/* Modo normal */
    cpu.r0i=2;		/* PC int. inicial */
    cpu.r[0]=0;		/* PC inicial */
    cpu.ir=0;		/* IR inicial (nop) */
    cpu.nciclos=0;	/* Contador de ciclos */
    cpu.ninstrs=0;	/* Contador de instrucciones */
    cpu.nnops=0;	/* Contador de nops */
    for (i=1;i<8;i++) cpu.r[i]=-1;
}

/*--------------------------------------------------------------------*/


void loadhex(char *fn)
{
    int i,dir,dato;
    char *p,*pp,buf[256];
    FILE *fp;
    
    if ((fp=fopen(fn,"r"))==NULL) exit(0);
    
    for (dir=0;;) {
        fgets(buf,255,fp);
	if (feof(fp)) break;
	p=buf;
	if (*p=='@') {
	   dir=strtol(++p,&pp,16);
	   continue;
	}
	
	MEM[dir]=strtol(p,&pp,16);
	dir++;
    }
    fclose(fp);
}

void alarma() // Se llama periodicamente cada 40 ms
{
}

void ctrlc()
{
    Trace=1;
}

#define FCLK 1.8432e6

main(int argc, char **argv)
{
	int i;	
	struct itimerval t0,t1;
        struct sigaction act0,act1;
          
	if (argc<2) {
	   printf("Uso: %s [opciones] fichero.hex\n",argv[0]);
	   printf("  opciones:\n");
	   printf("            -rr      R1 y R2 en vez de SP y LR\n");
	   printf("            -sp <n>  especifica el puntero de pila\n");
	   exit(0);
	}
	
	for (i=1;i<argc;i++) {
	    if (argv[i][0]!='-') loadhex(argv[i]); 
	    else {
	    	if (strcmp(argv[i],"-rr")==0) rgn=rgn0;
		if (strcmp(argv[i],"-sp")==0) spreg=atoi(argv[++i]);
	    }
	}
	for (i=0;i<80*24;i++) MEM[i+0x1000]=' ';  // CLS
	    
        t0.it_interval.tv_sec=0;
        t0.it_interval.tv_usec=40000;	// alarma cada 40 ms
        t0.it_value=t0.it_interval;

        act0.sa_handler=alarma;
        act0.sa_flags=SA_RESTART;

        sigaction(SIGALRM,&act0,&act1);
        setitimer(ITIMER_REAL,&t0,&t1);
	
	signal(SIGINT,ctrlc);

	reset_cpu();
	
	Trace=1;
	
	// 25 Interrupciones por segundo (T=40 ms)
	for (;;) {
	    for(i=((FCLK+12.5)/25-1);i>=0;i--) un_ciclo((i==0));
	    pause();
	}
}
