	org 0	; RESET vector
	
	jr	init
	xor	pc,pc,sp	; Init SP form PC (PC is the only known
				; register after reset. Verilog will keep
				; SP in unknown state otherwise)
	org 2	; IRQ vector
	
	reti
	nop

init:	
	xor	sp,sp,lr	; This init is required
	xor	sp,sp,r3	; for the Verilog simulator
	xor	sp,sp,r4	; only:
	xor	sp,sp,r5	; all registers have to be
	xor	sp,sp,r6	; computed from a known value,
	xor	sp,sp,r7	; in this case, SP

	ld	(pc),lr		; init video pointer
	word	pchar		; (not required for dummy
	ld	(pc),r3		; character out)
	word	4096
	st	r3,(lr)
	
start:
	ld	(pc),lr		; Fill sieve with ones
	word	array
	ld	(pc),r3
	word	128		; 2048/16
	ld	(pc),r4
	word	65535
	
l1:	st	r4,(lr)
	subq	r3,1,r3
	jr.nz	l1
	addq	lr,1,lr

	; r7 number
	
	ld	(pc),r7		; start with number 2
	word 	2	
mbuc:	ror	r7,r6		; compute pointer
	ror	r6,r6
	ror	r6,r6
	ror	r6,r6
	ld	(pc),lr
	word	4095
	and	r6,lr,r6
	ld	(pc),lr
	word	array
	add	r6,lr,r6
	ld	(pc),r4		; compute mask	
	word	15
	ld	(pc),r3
	word	1
	and	r4,r7,r4
	jr.z	l3
	nop
l2:	subq	r4,1,r4
	jr.nz	l2
	add	r3,r3,r3
	
l3:	ld	(r6),r4		; check bit
	and	r4,r3,r3
	jr.z	nextprime
	nop
	
	; Prime number: print it
	
	; divide by 10
	xor	r3,r3,r3	; remainder
	or	r7,r7,r6
	ld	(pc),r5		; divider
	word	10
	xor	r4,r4,r4	; number of digits

l35:	ld	(pc),lr		; loop counter
	word	16
	
	add	r6,r6,r6	; shift left
l4:	adc	r3,r3,r3
	sub	r3,r5,r3
	add.c	r3,r5,r3	; restore value
	addq.nc	r6,1,r6		; update quotient
l6:	subq	lr,1,lr
	jr.nz	l4
	add	r6,r6,r6	; shift left
	ror	r6,r6		; undo last shift

	ld	(pc),lr
	word	48		; ASCII '0'
	add	lr,r3,r3
	subq	sp,1,sp		; to stack
	st	r3,(sp)
	addq	r4,1,r4
	or	r6,r6,r6
	jr.nz	l35
	xor	r3,r3,r3
	
l7:	ld	(sp),r3		; from stack
	addq	pc,2,lr
	jr	putch		; print it
	addq	sp,1,sp
	subq	r4,1,r4
	jr.nz	l7
	nop
	
	ld	(pc),r3
	word	32		; space
	addq	pc,2,lr
	jr	putch		; print it
	nop

	; Now, mark multiples as not primes
	; r6 = index
	add	r7,r7,r6	; index=2*number
bucclr:	ld	(pc).lr		; compare with 2048
	word	2048
	sub	r6,lr,lr
	jr.nc	nextprime
	nop
	ror	r6,r5		; compute pointer
	ror	r5,r5
	ror	r5,r5
	ror	r5,r5
	ld	(pc),lr
	word	4095
	and	r5,lr,r5
	ld	(pc),lr
	word	array
	add	r5,lr,r5

	ld	(pc),r4		; compute mask	
	word	15
	ld	(pc),r3
	word	1
	and	r4,r6,r4
	jr.z	l9
	nop
l8:	subq	r4,1,r4
	jr.nz	l8
	add	r3,r3,r3
	
l9:	ld	(r5),r4		; clear bit
	not	r3,r3	
	and	r4,r3,r3
	st	r3,(r5)

	jr	bucclr
	add	r7,r6,r6	; add number to index


nextprime:
	ld	(pc),r6		; compare with max
	word	2047
	sub	r7,r6,r6
	jr.nz	mbuc
	addq	r7,1,r7		; next candidate

hang:	jr	hang
	nop		

;-------------------------------------
; routine to print R3 as an ASCII char

;putch:	or	lr,lr,pc	; return (dummy print)
;	nop

putch:	subq	sp,1,sp		; save registers
	st	lr,(sp)
	subq	sp,1,sp	
	st	r4,(sp)
	
	ld	(pc),lr		; load vmem pointer to r4
	word	pchar
	ld	(lr),r4
	st	r3,(r4)		; store char to vmem
	addq	r4,1,r4		; increment pointer
	st	r4,(lr)		; save pointer
	
	ld	(sp),r4		; restore registers
	addq	sp,1,sp
	ld	(sp),lr
	or	lr,lr,pc	; return
	addq	sp,1,sp

pchar:	word	4096	; Pointer to video memory
array:
